#!/usr/bin/env python

import argparse
import os
import os.path
import re
import shutil
import sys

import rpm
import yum

from update_package.common import *


class Package(object):
    """
    Object encapsulating the properties of an RPM
    """
    def __init__(self, name, epoch, version, release, buildhost):
        """
        Create an instance from rpmquery data
        """
        self.name = name
        self.epoch = epoch
        self.version = version
        self.release = release

        if re.match(r'[0-9a-f]{12}$', buildhost):
            self.category = 'local'
        elif 'centos' in buildhost:
            self.category = 'centos'
        else:
            self.category = 'other'

    def evr(self):
        """
        Return a string representation of the epoch, version, release
        """
        ret = ''
        if self.epoch != '0':
            ret = self.epoch+':'
        ret += "{}-{}".format(self.version, self.release)
        return ret

    def compare(self, other):
        """
        Return the operator that describes the relationship to `other`
        """
        val = rpm.labelCompare((self.epoch, self.version, self.release),
                               (other.epoch, other.version, other.release))
        if val > 0:
            return '>'
        elif val < 0:
            return '<'
        return '='


def parse_args_or_exit(argv=None):
    """
    Parse command line options
    """
    parser = argparse.ArgumentParser(description="Compare repositories/ISOs")
    parser.add_argument("repo_first", metavar="ISO",
                        help="First repository")
    parser.add_argument("repo_second", metavar="ISO",
                        help="Second repository")
    parser.add_argument("--category", dest='categories', action='append',
                        help="One or more of: 'local', 'centos', 'other'")

    return parser.parse_args(argv)


def find_all_repos():
    """
    Return the ids of all repositories configured
    """
    yum_base = yum.YumBase()
    return [r.id for r in yum_base.repos.findRepos('*')]


def setup_repo(repo, all_repos, work_dir):
    """
    Prepare a repo config for an ISO or directory
    """
    if repo in all_repos:
        return (repo, None)
    local_repo = fetch_remote_file(repo, work_dir)
    if os.path.isdir(local_repo):
        repo_dir = local_repo
    else:
        repo_dir = unpack_into(local_repo, work_dir)
    repo_id = repo.translate(None, '@~#/{}[]!"$%^&*()')
    config_dir = os.path.join(work_dir, 'yum.repos.d')
    if not os.path.exists(config_dir):
        os.mkdir(config_dir)
    config_fname = os.path.join(config_dir, repo_id+'.repo')
    with open(config_fname, 'w') as config_fh:
        print >>config_fh, """[{id}]
name={id}
baseurl=file://{dir}
gpgcheck=0
""".format(id=repo_id, dir=repo_dir)
    return (repo_id, config_dir)


def load_repo(id, config_dir, categories):
    """
    Return a dictionary of Packages in a repo, filtered by category
    """
    repo = {}
    format = '%{name} %{epoch} %{version} %{release} %{buildhost}'
    cmd = ['repoquery']
    if config_dir:
        cmd += ['--setopt', "reposdir="+config_dir]
    cmd += ['--disablerepo', '*', '--enablerepo', id]
    cmd += ['-q', '--qf', format, '*']
    output = subprocess.check_output(cmd)
    for line in output.split('\n'):
        if ' ' in line:
            values = line.split()
            p = Package(*values)
            if categories is None or p.category in categories:
                repo[values[0]] = p
    return repo


def summarize_content(first_repo, second_repo):
    """
    Summarize the package differences between two repos
    """
    first_pkgs = set(first_repo.keys())
    second_pkgs = set(second_repo.keys())

    only_in_first = first_pkgs-second_pkgs
    only_in_second = second_pkgs-first_pkgs

    if only_in_first:
        names_vers = ["{}-{}".format(n, first_repo[n].evr())
                      for n in only_in_first]
        print "\nPackages only in first repository:", ', '.join(names_vers)
    if only_in_second:
        names_vers = ["{}-{}".format(n, second_repo[n].evr())
                      for n in only_in_second]
        print "\nPackages only in second repository:", ', '.join(names_vers)


def compare_versions(first_repo, second_repo):
    """
    Report the differences in EVR between the packages common to two repos
    """
    first_pkgs = set(first_repo.keys())
    second_pkgs = set(second_repo.keys())
    common_pkgs = sorted(first_pkgs & second_pkgs)

    diffs = [p for p in common_pkgs
             if first_repo[p].evr() != second_repo[p].evr()]
    if diffs:
        print "\nPackage version differences:\n"
        for d in diffs:
            first = first_repo[d]
            second = second_repo[d]
            print "\t{:<32}  {} {} {}".format(d, first.evr(),
                                              first.compare(second),
                                              second.evr())


def main(argv):
    """
    Compare the contents of two repositories
    """
    args = parse_args_or_exit(argv)
    work_dir = tempfile.mkdtemp(prefix='build-update-')
    all_repos = find_all_repos()
    try:
        first_id, first_dir = setup_repo(args.repo_first, all_repos, work_dir)
        first_repo = load_repo(first_id, first_dir, args.categories)
        second_id, second_dir = setup_repo(args.repo_second, all_repos,
                                           work_dir)
        second_repo = load_repo(second_id, second_dir, args.categories)

        summarize_content(first_repo, second_repo)
        compare_versions(first_repo, second_repo)
    finally:
        shutil.rmtree(work_dir, ignore_errors=True)


def _main():
    """
    Entry point for setuptools CLI wrapper
    """
    main(sys.argv[1:])


# Entry point when run directly
if __name__ == "__main__":
    _main()
