#!/usr/env python

import argparse
import glob
import os
import os.path
import shutil
import subprocess
import sys
import tempfile

from update_package.buildcommon import *
from update_package.common import *


def check_uuid(uuid):
    """
    Validate UUID
    """
    if not valid_uuid(uuid):
        sys.exit("%s: invalid UUID: %s" % (sys.argv[0], uuid))
    return uuid


def check_label(label):
    """
    Validate update label
    """
    if not valid_label(label):
        sys.exit("%s: invalid label: %s" % (sys.argv[0], label))
    return label


def check_guidance(guidance):
    """
    Validate after-apply guidance
    """
    if not valid_guidance(guidance):
        sys.exit("%s: invalid guidance: %s" % (sys.argv[0], guidance))
    return guidance


def parse_args_or_exit(argv=None):
    """
    Parse command line options
    """
    parser = argparse.ArgumentParser(description="Build update media")
    parser.add_argument("rpms", metavar="RPM", nargs="+",
                        help="Packages to include")
    parser.add_argument("--uuid", type=check_uuid, required=True,
                        help="Update unique identifier")
    parser.add_argument("-l", "--label", type=check_label, required=True,
                        help="Update textual label")
    parser.add_argument("-v", "--version", required=True,
                        help="Update version")
    parser.add_argument("-d", "--description", required=True,
                        help="Update description")
    parser.add_argument("-o", "--output", metavar="ISO", required=True,
                        help="Output ISO file")
    parser.add_argument("-k", "--key", metavar="UID", default=None,
                        help="UID of gpg key")
    parser.add_argument("--keyfile", default=None,
                        help="File name of gpg key")
    parser.add_argument("--base-requires", default=None,
                        help="Product version dependency")
    parser.add_argument("--after-apply", type=check_guidance, default=None,
                        help="Post apply guidance")
    parser.add_argument("--precheck", default=[], action='append',
                        help="Precheck script")
    parser.add_argument("--remove", default=None,
                        help="Uninstall script")
    parser.add_argument('--no-passphrase', action='store_false',
                        dest="prompt_pass", help="Disable passphrase prompt")
    parser.add_argument("--groupfile", default=None,
                        help="Group information")
    parser.add_argument('--enforce-homogeneity', action='store_true',
                        help="Include this update in pool join checking")

    return parser.parse_args(argv)


def main(argv):
    """
    Create and sign constituent parts then assemble ISO
    """
    args = parse_args_or_exit(argv)
    passphrase = None
    control_pkg = None
    if args.key and args.prompt_pass:
        passphrase = get_gpg_key_pass(args.key)
    work_dir = tempfile.mkdtemp(prefix='build-update-')
    try:
        os.mkdir(os.path.join(work_dir, 'root'))
        os.mkdir(os.path.join(work_dir, 'SOURCES'))
        has_control = args.base_requires or args.precheck or args.remove
        if has_control:
            control_pkg = build_control_rpm(work_dir, args.uuid, args.label,
                                            args.version, args.base_requires,
                                            args.precheck, remove=args.remove)
        local_rpms = [fetch_remote_file(rpm, work_dir) for rpm in args.rpms]
        size = calc_install_size(local_rpms)
        create_update_doc(os.path.join(work_dir, 'stampfile.xml'), args.uuid,
                          args.label, args.version, args.description,
                          args.after_apply, None, control_pkg, size,
                          enforce_homogeneity=args.enforce_homogeneity)
        build_update_rpm(work_dir, args.uuid, args.label, args.version,
                         args.description, local_rpms, args.after_apply)
        create_update_doc(os.path.join(work_dir, 'root/update.xml'), args.uuid,
                          args.label, args.version, args.description,
                          args.after_apply, args.keyfile, control_pkg, size,
                          enforce_homogeneity=args.enforce_homogeneity)
        copy_packages(work_dir, local_rpms)
        if args.key:
            sign_update_doc(work_dir, args.key, passphrase)
            packages = glob.glob(os.path.join(work_dir, "root/Packages/*.rpm"))
            sign_packages(work_dir, args.key, passphrase, packages)
        if args.groupfile:
            groupfile = args.groupfile
        else:
            groupfile = group_file(work_dir, args.label)
        create_repo(work_dir, args.key, passphrase, groupfile)
        build_iso(work_dir, args.output, args.uuid, args.label)
    finally:
        shutil.rmtree(work_dir, ignore_errors=True)


def _main():
    """
    Entry point for setuptools CLI wrapper
    """
    main(sys.argv[1:])


# Entry point when run directly
if __name__ == "__main__":
    _main()
