from __future__ import print_function
import glob
import json
import os
import os.path
import re
import shutil
import subprocess
import tarfile
import xml.dom.minidom
import xml.etree.ElementTree as ET
from subprocess import check_output, CalledProcessError
from StringIO import StringIO

from pkg_resources import resource_filename

from update_package.common import *


def calc_install_size(rpms):
    """
    Calculate the installation-size
    """
    total_size = 0

    for rpm in rpms:
        size, = rpm_query(rpm, 'size')
        total_size += int(size)

    return total_size


def create_relationship_element(element, relation, uuid_labels):
    for each_uuid_label in uuid_labels:
        uuid, label = each_uuid_label.split(":", 1)
        relation_element = element.ownerDocument.createElement(relation)
        if label:
            relation_element.setAttribute('name-label', label)
        if uuid:
            relation_element.setAttribute('uuid', uuid)
        element.appendChild(relation_element)


def create_update_doc(filename, uuid, label, version, description, after_apply,
                      keyfile, control_pkg, size, guidance_mandatory=None,
                      requires=None, conflicts=None, rollsup=None,
                      properties=None, build_number=None,
                      enforce_homogeneity=False, rolled_up_by=None,
                      follows=None, stream=None):
    """
    Create XML document describing update
    """
    dom = xml.dom.minidom.getDOMImplementation()
    doc = dom.createDocument(None, "update", None)
    top = doc.documentElement
    top.setAttribute('name-label', label)
    top.setAttribute('version', version)
    top.setAttribute('uuid', uuid)
    top.setAttribute('installation-size', str(size))
    if after_apply:
        top.setAttribute('after-apply-guidance', after_apply)
        if guidance_mandatory:
            top.setAttribute('guidance-mandatory', "true")
    if keyfile:
        top.setAttribute('key', os.path.basename(keyfile))
    if control_pkg:
        top.setAttribute('control', control_pkg)
    desc = doc.createElement("name-description")
    desc.appendChild(doc.createTextNode(description))
    top.appendChild(desc)
    if requires:
        create_relationship_element(top, 'requires', requires)
    if conflicts:
        create_relationship_element(top, 'conflicts', conflicts)
    if rollsup:
        create_relationship_element(top, 'rollsup', rollsup)
    if rolled_up_by:
        # rolled_up_by isn't a list
        create_relationship_element(top, 'rolled-up-by', [rolled_up_by])
    if properties:
        top.setAttribute('properties', properties)
    if build_number:
        top.setAttribute('build-number', build_number)
    top.setAttribute('enforce-homogeneity', "true" if enforce_homogeneity
                     else "false")
    if follows:
        follows_no_uuid = [ ':' + label for label in follows ]
        create_relationship_element(top, 'follows', follows_no_uuid)
    if stream:
        top.setAttribute('update-stream', stream)

    with open(filename, "w") as fileh:
        top.writexml(fileh, addindent="  ", newl='\n')


def relationship_line(tag, uuid_label, prefix="update"):
    """
    Split a UUID:label and form a tag and value
    """
    _, label = uuid_label.split(':', 1)
    return "%s: %s-%s\n" % (tag, prefix, label)


prep_install_frag = """
%prep
%setup -c

%install
for f in $(find . -name '*.xml'); do
  uuid=$(xmllint --xpath "string(//update/@uuid)" $f)
  install -m644 -D $f $RPM_BUILD_ROOT/var/update/applied/$uuid
done
"""


def build_update_rpm(work_dir, uuid, label, version, description, rpms,
                     after_apply, rollsup=None, source=None,
                     build_number=None):
    """
    Create update RPM to deploy update payload
    """
    # create files that are included by the spec file
    if source:
        shutil.copy(source, os.path.join(work_dir, "SOURCES"))
        with open(os.path.join(work_dir, "source0"), "w") as fileh:
            fileh.write("Source0: %s\n" % os.path.basename(source))
        with open(os.path.join(work_dir, "prep_install"), "w") as fileh:
            fileh.write(prep_install_frag)
    with open(os.path.join(work_dir, "description"), "w") as fileh:
        fileh.write(description + '\n')
    no_livepatch = True
    with open(os.path.join(work_dir, "requires"), "w") as fileh:
        for rpm in rpms:
            epoch, = rpm_query(rpm, 'epoch')
            if epoch == '(none)':
                dep_str = '%{name} >= %{version}-%{release}'
            else:
                dep_str = '%{name} >= %{epoch}:%{version}-%{release}'
            rpmdep = subprocess.check_output(['rpmquery', '--nosignature',
                                              '--qf', dep_str, '-p', rpm])
            fileh.write("Requires: %s\n" % rpmdep)
            rpmbase = os.path.basename(rpm)
            if rpmbase.startswith('kernel-livepatch_') or \
               rpmbase.startswith('xen-livepatch_'):
                no_livepatch = False
    if rollsup:
        with open(os.path.join(work_dir, "obsoletes"), "w") as fileh:
            for rollup in rollsup:
                fileh.write(relationship_line("Obsoletes", rollup))
                fileh.write(relationship_line("Obsoletes", rollup, "control"))
        if source:
            with open(os.path.join(work_dir, "provides"), "w") as fileh:
                for rollup in rollsup:
                    fileh.write(relationship_line("Provides", rollup))

    if no_livepatch and isinstance(after_apply, basestring) and \
       'restartHost' in after_apply:
        with open(os.path.join(work_dir, "posttrans"), "w") as fileh:
            fileh.write("echo %s >>/run/reboot-required.hfxs\n" % uuid)

    if build_number is None:
        build_number = ''
    subprocess.check_output(['rpmbuild', '--define', '_topdir %s' % work_dir,
                             '--define', 'label %s' % label,
                             '--define', 'version %s' % version,
                             '--define', 'uuid %s' % uuid,
                             '--define', 'build_number %s' % build_number,
                             '-bb', resource_filename('update_package',
                                                      'update.spec')])


def interpreter(script):
    """
    Determine the interpreter used by a script
    """
    with open(script) as fileh:
        line = fileh.readline().rstrip()
    if line.startswith('#!'):
        line = line[2:].strip()
    args = line.split()
    if args[0].endswith('/env'):
        return subprocess.check_output(['which', args[1]])
    else:
        return args[0]

multi_precheck_script = """#!/bin/sh
set -e

dir=$(dirname $0)/precheck.d
for precheck in $dir/S*; do
  $precheck
done
"""


def build_control_rpm(work_dir, uuid, label, version, base_requires,
                      prechecks=[], requires=None, conflicts=None,
                      remove=None):
    """
    Create control RPM to deploy scripts
    """
    # create source tarball
    ctrl_source = os.path.join(work_dir, "SOURCES",
                               "control-%s-source.tar" % label)
    with tarfile.open(ctrl_source, "w") as tar:
        if len(prechecks) == 1:
            tar.add(prechecks[0], arcname="precheck")
        elif prechecks:
            # multiple precheck scripts
            entry = tarfile.TarInfo("precheck")
            entry.size = len(multi_precheck_script)
            tar.addfile(entry, StringIO(multi_precheck_script))
            sequence = 0
            for precheck in prechecks:
                tar.add(precheck, arcname="precheck.d/S%02d%s" % (sequence, os.path.basename(precheck)))
                sequence += 1
        if remove:
            tar.add(remove, arcname="remove")

    # create list of Requires: and Conflicts
    with open(os.path.join(work_dir, "requires-conflicts"), "w") as fileh:
        if base_requires:
            fileh.write("Requires: %s\n" % base_requires)
        if requires:
            for require in requires:
                fileh.write(relationship_line("Requires", require))
        if conflicts:
            for conflict in conflicts:
                fileh.write(relationship_line("Conflicts", conflict))

    # create script list include by the spec file
    with open(os.path.join(work_dir, "control-files"), "w") as fileh:
        if prechecks:
            fileh.write("/var/update/%s/precheck\n" % uuid)
        if len(prechecks) > 1:
            fileh.write("/var/update/%s/precheck.d\n" % uuid)
        if remove:
            fileh.write("/var/update/%s/remove\n" % uuid)

    subprocess.check_output(['rpmbuild', '--define', '_topdir %s' % work_dir,
                             '--define', 'label %s' % label,
                             '--define', 'version %s' % version,
                             '--define', 'uuid %s' % uuid,
                             '-bb', resource_filename('update_package',
                                                      'control.spec')])
    return "control-" + label


def copy_packages(work_dir, rpms):
    """
    Copy payload and any additional RPMS
    """
    pkg_dir = os.path.join(work_dir, 'root/Packages')
    os.mkdir(pkg_dir)
    extra_pkgs = glob.glob(os.path.join(work_dir, 'RPMS/noarch/*.rpm'))
    for rpm in rpms + extra_pkgs:
        shutil.copy2(rpm, pkg_dir)


def driver_disk_group_file(filename, label, driver_packages=[]):
    """
    Create a standard group file for driver disk updates
    """
    packages = ''
    for driver in driver_packages:
        packages += "\n      <packagereq type=\"mandatory\">%s</packagereq>" % driver

    with open(filename, "w") as fileh:
        fileh.write("""<!DOCTYPE comps PUBLIC "-//Red Hat, \
Inc.//DTD Comps info//EN" "comps.dtd"> 
<comps>
  <group>
   <id>update</id>
   <default>False</default>
   <uservisible>True</uservisible>
   <name>%(label)s</name>
   <description></description>
    <packagelist>
      <packagereq type="mandatory">update-%(label)s</packagereq>
    </packagelist>
  </group>
  <group>
   <id>drivers</id>
   <default>False</default>
   <uservisible>True</uservisible>
   <name>%(label)s device drivers</name>
   <description></description>
    <packagelist>%(packages)s
    </packagelist>
  </group>
</comps>
""" % {'label': label, 'packages': packages})

def group_file(work_dir, label):
    """
    Create a standard group file
    """
    groupfile = os.path.join(work_dir, "groups.xml")
    with open(groupfile, "w") as fileh:
        fileh.write("""<!DOCTYPE comps PUBLIC "-//Red Hat, \
Inc.//DTD Comps info//EN" "comps.dtd">
<comps>
  <group>
   <id>update</id>
   <default>False</default>
   <uservisible>True</uservisible>
   <name>%(label)s</name>
   <description></description>
    <packagelist>
      <packagereq type="mandatory">update-%(label)s</packagereq>
    </packagelist>
  </group>
</comps>
""" % {'label': label})
    return groupfile


def read_group_file(repo):
    """
    Locate and read the group file for a repository
    """
    group_path = None
    repomd = xml.dom.minidom.parse(os.path.join(repo, 'repodata/repomd.xml'))
    for data_el in repomd.getElementsByTagName('data'):
        if data_el.getAttribute('type') == 'group':
            location_el = data_el.getElementsByTagName('location')
            group_path = location_el[0].getAttribute('href')
            break
    if group_path:
        return ET.parse(os.path.join(repo, group_path))
    return None


def packages_in_group(doc, group_name):
    """
    Return the list of package names in a group
    """
    root = doc.getroot()
    packages = None
    for group in root.iter('group'):
        ident = group.find('id')
        if ident.text == group_name:
            packages = [ x.text for x in group.iter('packagereq') ]
            break

    if not packages:
        raise LookupError("Group %s not found in doc %s" % (group_name, doc))
    return packages


def create_yum_repo_conf(iso_paths, repo_path):
    """
    Create an RPM merge.repo file containing a repository for each input ISO
    """
    with open(repo_path, 'w') as repofile:

        # Write a main section
        print("[main]", file=repofile)
        print("gpgcheck = 0", file=repofile)
        print("assumeyes=  1", file=repofile)
        print("installonlypkgs = ", file=repofile)
        print("", file=repofile) # print automatically adds a newline

        for iso_path in iso_paths:
            iso_name = os.path.basename(iso_path)
            
            print("[%s]" % iso_name, file=repofile)
            print("name = %s" % iso_name, file=repofile)
            
            abs_path = os.path.abspath(iso_path) # Ensure a leading /
            rpm_file_path = "file://%s/" % abs_path # RPM file paths must start with ///
            
            print("baseurl = %s" % rpm_file_path, file=repofile)
            print("", file=repofile) # print automatically adds a newline


def valid_uuid(uuid):
    """
    Validate UUID
    """
    if re.match(r'[\da-f]{8}-[\da-f]{4}-[\da-f]{4}-[\da-f]{4}-[\da-f]{12}$',
                uuid):
        return True
    return False


def valid_label(label):
    """
    Validate update label
    """
    if re.match(r'[\w-]+$', label):
        return True
    return False


def valid_uuid_label(uuid_label):
    if ':' not in uuid_label:
        return False
    uuid, label = uuid_label.split(':', 1)
    return valid_uuid(uuid) and valid_label(label)


def valid_guidance(guidance):
    """
    Validate after-apply guidance
    """
    valid = ("restartXAPI", "restartHost", "restartPV", "restartHVM")
    for g in guidance.split(','):
        if g not in valid:
            return False
    return True


def check_output_wrapper(cmd, **kwargs):
    """
    Wrap check_output to output stdout on error
    Notable commands that need this: yumdownloader
    """
    try:
        return subprocess.check_output(cmd, **kwargs)
    except CalledProcessError as err:
        print(err.output) # stdout, some commands write error messages here
        raise err

