import getpass
import glob
import hashlib
import json
import os
import os.path
import shutil
import subprocess
import sys
import tempfile
import urllib
import urlparse
import xml.dom.minidom

import libarchive
import rpm


def artifactory_auth():
    auth = None
    config_path = os.path.expanduser('~/.artifactory_python.cfg')
    p = ConfigParser.ConfigParser()
    p.read(config_path)

    for section in p.sections():
        username = p.get(section, 'username') if p.has_option(section, 'username') else None
        password = p.get(section, 'password') if p.has_option(section, 'password') else None

        auth = "%s:%s" % (username, password)

    return auth


def copy_remote_file(src, dest):
    """
    Copy src to dest fetching the file if src is a remote URL
    """
    url = urlparse.urlparse(src)
    if url.scheme == '':
        # local path
        shutil.copy2(src, dest)
    else:
        print "Fetching %s..." % src
        cmd = ['curl', '-f', '-L', src, '-o', dest]
        auth = artifactory_auth()
        if auth:
            cmd += ['-u', auth]
        subprocess.check_call(cmd)


def fetch_remote_file(name, work_dir=None):
    """
    If name is a remote URL fetch the file to a local directory
    """
    url = urlparse.urlparse(name)
    if url.scheme == '':
        # local path
        return name

    if work_dir is None:
        work_dir = tempfile.mkdtemp(prefix='fetch-file-')
    local_name = os.path.join(work_dir, os.path.basename(url.path))
    copy_remote_file(name, local_name)

    return local_name


def replace_path(url, new_path):
    """
    Replace the path part of a URL
    """
    url = urlparse.urlparse(url)
    url = url._replace(path = new_path)
    return urlparse.urlunparse(url)
 

def search_pattern(pattern):
    """
    Get a list of artifactory URLs matching a given pattern
    """
    url = urlparse.urlparse(pattern)
    if not url.path or url.path == '/':
        raise ValueError('Pattern must include a URL path')

    path = url.path.lstrip('/') # Remove leading /
    try:
        repo_key, rest = path.split('/', 1) # repo_key is eg. xs-local-assembly
    except ValueError:
        raise ValueError('Pattern must include an Artifactory repo (eg. xs-local-build) and a path')

    # Pattern param format for Artifactory is repo_key:path/to/dir/*.ext
    params = {'pattern': repo_key + ':' + rest} 
    api_path = '/api/search/pattern'
    url = url._replace(path=api_path)
    url = url._replace(query=urllib.urlencode(params))
    url = urlparse.urlunparse(url)

    cmd = ['curl', '-f', '--silent', url]
    auth = artifactory_auth()
    if auth:
        cmd += ['-u', auth]

    out = subprocess.check_output(cmd)
    jdata = json.loads(out)

    # jdata['files'] is a list of matches, but doesn't include repo_key
    # eg. packaging-tools/stable/1.x/RPMS/noarch/update-package-1.9.11-1.noarch.rpm
    # Join it with repo_key in the artifactory URL
    return [ replace_path(pattern, os.path.join(repo_key, file)) for file in jdata['files']]


def is_url(location):
    res = urlparse.urlparse(location)
    return all([res.scheme, res.netloc, res.path])


import ConfigParser


def deploy_file(local_path, repo_path):
    """
    Upload a file to an artifactory repository
    """
    remote_path = os.path.join(repo_path, os.path.basename(local_path))
    deploy_to_filename(local_path, remote_path)


def deploy_to_filename(local_path, repo_file):
    """
    Upload a file to an artifactory location
    """
    print "Computing hashes for " + local_path
    md5sum = hashlib.md5()
    sha1sum = hashlib.sha1()
    with open(local_path) as fileh:
        while True:
            block = fileh.read(16 * 1024 * 1024)
            if len(block) == 0:
                break
            md5sum.update(block)
            sha1sum.update(block)
    md5digest = md5sum.hexdigest()
    sha1digest = sha1sum.hexdigest()

    auth = artifactory_auth()
    cmd = ['curl', '-f', '-i', '-X', 'PUT',
           '-H', "X-Checksum-Md5: " + md5digest,
           '-H', "X-Checksum-Sha1: " + sha1digest, '-u', auth, '-T',
           local_path, repo_file]
    subprocess.check_call(cmd)


def in_place_copy(src_url, dst_url, dst_name=""):
    """
    Perform an in-place copy using the artifactory API
    """

    # Identify the artifactory URL and the source and dest paths
    src = urlparse.urlparse(src_url)
    dst = urlparse.urlparse(dst_url)

    # Normalize ports
    default_port = {'http': 80, 'https': 443}
    src_port = src.port if src.port else default_port.get(src.scheme)
    dst_port = dst.port if dst.port else default_port.get(dst.scheme)

    if src.hostname != dst.hostname or src_port != dst_port:
        raise ValueError("Source and destination must both be on the same Artifactory instance")

    src_path = os.path.normpath(src.path)
    if not dst_name:
        dst_name = os.path.basename(src_path)
    dst_path = os.path.join(dst.path, dst_name)
    if src_path == dst_path:
        raise ValueError("Source and destination are the same")

    uri = "https://%s/api/copy%s?to=%s" % (src.netloc, src_path, dst_path)

    auth = artifactory_auth()
    cmd = ['curl', '-f', '-i', '-X', 'POST', '-u', auth, uri]
    subprocess.check_call(cmd)


def call_stderr(*args):
    """
    Call a command, return its exit code and any error output
    """
    proc = subprocess.Popen(*args, stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT)
    out, _ = proc.communicate()
    return proc.returncode, out


def unpack_archive(archive_name, unpack_dir):
    """
    Unpack ISO image or RPM into a directory
    """
    iso_path = os.path.abspath(archive_name)
    cwd = os.getcwd()
    os.chdir(unpack_dir)
    libarchive.extract_file(iso_path)
    subprocess.check_call(['chmod', '-R', 'u+w', '.'])
    os.chdir(cwd)


def unpack_into(iso_path, work_dir):
    """
    Unpack a (possibly remote) ISO into an empty directory
    """
    base = os.path.basename(iso_path)
    out_dir = os.path.join(work_dir, 'unpack', base)
    os.makedirs(out_dir)
    print("Unpacking " + base)
    unpack_archive(fetch_remote_file(iso_path, work_dir), out_dir)
    return out_dir


def read_volume_desc(iso_name):
    """
    Read volume label and UUID from ISO image
    """
    desc = {}
    out = subprocess.check_output(['isoinfo', '-d', '-i', iso_name])
    for line in out.split('\n'):
        if ': ' in line:
            k, v = line.split(': ', 1)
            desc[k] = v
    uuid = desc['Application id']
    if len(uuid) != 36:
        uuid = None
    label = desc['Volume id']
    return uuid, label


def gpg(passphrase, args):
    """
    Invoke gpg, securely passing any passphrase via stdin
    """
    if passphrase:
        passphrase += "\n"
    cmd = [os.environ.get('UPDATE_GPG_BIN', 'gpg'), '--batch', '--yes',
           '--passphrase-fd', '0']
    proc = subprocess.Popen(cmd + args, stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    out, err = proc.communicate(passphrase)
    if proc.returncode != 0:
        sys.exit("%s: gpg operation failed, %s" % (sys.argv[0], err))
    return out


def sign_packages(work_dir, key, passphrase, packages):
    """
    Sign RPMs with a GnuPG key
    """
    rpm.addMacro('_gpg_name', key)
    if 'UPDATE_GPG_BIN' in os.environ:
        rpm.addMacro('__gpg', os.environ['UPDATE_GPG_BIN'])

    for package in packages:
        if passphrase:
            ret = rpm.addSign(package, passphrase)
        else:
            # Older versions of rpm always expect two arguments to addSign(),
            # where an empty string is given for no passphrase.
            # Newer versions hang if an empty string is used when a key has
            # no passphrase.
            try:
                ret = rpm.addSign(package)
            except TypeError:
                # Two args required, fall back to an empty string
                ret = rpm.addSign(package, '')
                passphrase = ''
        if not ret:
            raise ValueError('Failed to sign ' + package)


def sign_update_doc(work_dir, key, passphrase):
    """
    Sign the update XML document
    """
    gpg(passphrase, ['--default-key', key, '--detach-sign', '--armor',
                     os.path.join(work_dir, 'root/update.xml')])


def copy_group_file(work_dir):
    """
    Locate the group definition and copy it
    """
    repomd = os.path.join(work_dir, 'root/repodata/repomd.xml')
    if not os.path.exists(repomd):
        return None
    doc = xml.dom.minidom.parse(repomd)
    top = doc.documentElement
    dest_file = None

    for el in top.getElementsByTagName("data"):
        if el.getAttribute("type") == "group":
            loc = el.getElementsByTagName("location")
            group_path = loc[0].getAttribute("href")
            src_file = os.path.join(work_dir, 'root', group_path)
            dest_file = os.path.join(work_dir, 'groups.xml')
            shutil.copyfile(src_file, dest_file)
            break

    return dest_file


def create_repo(work_dir, key=None, passphrase=None, groupfile=None,
                pkgdir=None):
    """
    Generate repository metadata and sign it
    """
    root_dir = os.path.join(work_dir, 'root')
    cmd = ['createrepo']
    if pkgdir:
        xglobs = [os.path.basename(xdir) + '/*' for xdir in
                  glob.glob(root_dir + '/*') if os.path.basename(xdir) !=
                  pkgdir and os.path.isdir(xdir)]
        for xglob in xglobs:
            cmd += ['-x', xglob]
    if groupfile:
        cmd += ['--groupfile', os.path.abspath(groupfile)]
    cmd.append('.')
    subprocess.check_call(cmd, cwd=os.path.join(work_dir, 'root'))
    if key:
        gpg(passphrase, ['--default-key', key, '--detach-sign', '--armor',
                         os.path.join(work_dir, 'root/repodata/repomd.xml')])


def build_iso(work_dir, iso_name, uuid, label):
    """
    Create ISO image containing update
    """
    subprocess.check_call(['genisoimage', '-J', '-joliet-long', '-R',
                           '-V', label, '-A', uuid, '-o', iso_name,
                           os.path.join(work_dir, 'root')])


def build_bootable_iso(work_dir, iso_name, uuid, label):
    """
    Create bootable ISO media image
    """
    subprocess.check_call(['genisoimage', '-J', '-r', '-V', label, '-A', uuid,
                           '-joliet-long',
                           '-b', 'boot/isolinux/isolinux.bin',
                           '-c', 'boot/isolinux/boot.cat',
                           '-no-emul-boot', '-boot-load-size', '4',
                           '-boot-info-table', '-eltorito-alt-boot',
                           '-e', 'boot/efiboot.img', '-no-emul-boot',
                           '-o', iso_name, os.path.join(work_dir, 'root')])
    subprocess.check_call(['isohybrid', '-u', iso_name])


def read_update_doc(filename):
    """
    Read name label and UUID from update document
    """
    doc = xml.dom.minidom.parse(filename)
    top = doc.documentElement
    return (top.getAttribute('uuid'), top.getAttribute('name-label'))


def get_gpg_key_pass(key_id):
    """
    Obtain the password for the specified GPG key
    """
    # Default the display string to the ID in case we can't do better
    key_str = key_id
    keydetails = gpg(None, ['--batch', '--with-colons', '-k', key_id])
    for line in keydetails.split('\n'):
        split_line = line.split(':')
        if split_line[0] == 'uid':
            key_str = split_line[9]
    return getpass.getpass("Enter passphrase for %s: " % key_str)


def rpm_query(path, *tags):
    """
    Query the RPM database (or a package), returning a list of values for a
    given set of tags
    """
    sep = '\001'
    fmt = sep.join(["%%{%s}" % t for t in tags])
    cmd = ['rpmquery', '--nosignature', '--qf', fmt]
    if path.endswith('.rpm'):
        cmd.append('-p')
    cmd.append(path)
    return subprocess.check_output(cmd).split(sep)
