import argparse
import os
import os.path
import shutil
import subprocess
import sys
import tempfile

import rpm
from urlgrabber import urlopen


# The following packages are unconditionally excluded
package_blacklist = ('lpe', 'xenserver-ddk-files', 'xenserver-vnc-control')


def make_repo_args(repos):
    repo_args = ['--disablerepo', '*']
    for repo in repos:
        repo_args += ['--enablerepo', repo]
    return repo_args


def download(url, fname):
    """
    Write the contents of a URL to a file
    """
    fh_in = urlopen(url)
    try:
        with open(fname, 'wb') as fh_out:
            shutil.copyfileobj(fh_in, fh_out)
    finally:
        fh_in.close()


def fetch_src(repos, pkgs, output_dir):
    """
    Search a list of repositories for each source package related to binary
    package.

    Returns: `found` - a dictionary of name: file path
             `not_found` - a list of packages with no source package
    """
    found = {}
    not_found = []
    args = ['--destdir', output_dir, '--source', '--url']
    repo_args = make_repo_args(repos)

    print "Searching for %s in %s" % (' '.join(pkgs), ' '.join(repos))
    for pkg in pkgs:
        # attempt to determine the file name of the source package
        tmp_dir = tempfile.mkdtemp(prefix='yum-',
                                   dir=os.environ.get('TMPDIR', '/tmp'))
        env = os.environ.copy()
        env['TMPDIR'] = tmp_dir
        cmd = ['yumdownloader']+repo_args+args+[pkg]
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env)
        output, _ = p.communicate()
        if p.returncode == 0:
            url = output.strip().split('\n')[-1]
            fname = os.path.join(output_dir, os.path.basename(url))
            download(url, fname)
            found[pkg] = fname
        else:
            not_found.append(pkg)
        shutil.rmtree(tmp_dir, ignore_errors=True)
    return found, not_found


def requires(srpms):
    """
    For a source package return its `Requires`, this being the `BuildRequires`
    within the spec file.
    """
    deps = set()
    ts = rpm.TransactionSet()
    ts.setVSFlags((rpm._RPMVSF_NOSIGNATURES | rpm._RPMVSF_NODIGESTS))
    for srpm in srpms:
        with open(srpm) as fh:
            header = ts.hdrFromFdno(fh.fileno())
        deps |= {d for d in header[rpm.RPMTAG_REQUIRES]
                 if not any(('(' in d, '/' in d))}
    return deps


def fetch_deps(repos, srpms, seen, output_dir):
    """
    For a set of source package recursively retrieve the packages required
    to build it.
    """
    # get all build dependencies
    deps = requires(srpms)
    # remove packages we have already seen or are in the blacklist
    new_deps = deps - set(seen.keys()) - set(package_blacklist)
    # fetch the source packages
    found, _ = fetch_src(repos, new_deps, output_dir)
    # remove filenames we already have seen
    srpms = set(found.values())-set(seen.values())
    if srpms:
        seen.update(found)
        # recurse for remaining source packages
        fetch_deps(repos, srpms, seen, output_dir)


def present_pkgs(repos, pkgs):
    """
    Return the set of packages actually present within a list of repos.
    """
    repo_args = make_repo_args(repos)
    tmp_dir = tempfile.mkdtemp(prefix='yum-',
                               dir=os.environ.get('TMPDIR', '/tmp'))
    try:
        env = os.environ.copy()
        env['TMPDIR'] = tmp_dir
        cmd = ['repoquery', '--qf', '%{name}']+repo_args+pkgs
        output = subprocess.check_output(cmd, env=env)
    finally:
        shutil.rmtree(tmp_dir, ignore_errors=True)
    return set(output.strip().split())


def parse_args_or_exit(argv=None):
    """
    Parse command line options
    """
    parser = argparse.ArgumentParser(description="Download source packages")
    parser.add_argument("build_repos", metavar="repo [ repo...]",
                        help="list of build repositories"),
    parser.add_argument("repos", metavar="repo [ repo...]",
                        help="list of all repositories"),
    parser.add_argument("pkgs", metavar="PKG", nargs="+",
                        help="Packages to resolve"),
    parser.add_argument("-o", "--output", metavar="DIR", required=True,
                        help="Output directory")
    parser.add_argument("-x", "--no-sources", metavar="repo [repo..]",
                        help="List of repositories without source packages")

    return parser.parse_args(argv)


def main(argv):
    """
    Download the source packages (including build dependencies)
    """
    args = parse_args_or_exit(argv)
    build_repos = args.build_repos.split()
    repos = args.repos.split()

    pkgs = []
    for pkg in args.pkgs:
        if pkg in package_blacklist:
            print "Excluding "+pkg
        else:
            pkgs.append(pkg)

    # first look for the source in the build repositories
    found, not_found = fetch_src(build_repos, pkgs, args.output)
    if not_found:
        # next look in the complete repository set
        _, not_found = fetch_src(repos, not_found, args.output)
        if not_found:
            # no source package present anywhere

            if args.no_sources:
                # find which were from the repositories without sources
                no_src_repos = args.no_sources.split()
                expected = present_pkgs(no_src_repos, not_found)
                print "Ignoring %s, binaries from no source repositories" % ' '.join(expected)
                unexpected = set(not_found)-expected
                if unexpected:
                    sys.exit("Missing source for "+' '.join(unexpected))
            else:
                sys.exit("Missing source for "+' '.join(not_found))

    fetch_deps(build_repos, found.values(), found, args.output)


def _main():
    """
    Entry point for setuptools CLI wrapper
    """
    main(sys.argv[1:])


# Entry point when run directly
if __name__ == "__main__":
    _main()
