#!/usr/env python

import argparse
import glob
import os
import shutil
import sys
import subprocess
import tempfile
import xml.dom.minidom
import ConfigParser

from update_package.buildcommon import *
from update_package.common import *


def unpack_iso(iso_name):
    """
    Unpack ISO image into working directory
    """
    work_dir = tempfile.mkdtemp(prefix='repack-update-')
    unpack_dir = os.path.join(work_dir, 'root')
    os.mkdir(unpack_dir)
    unpack_archive(iso_name, unpack_dir)
    return work_dir


def edit_update_doc(filename, config, uuid):
    """
    Replace description in update document
    """
    doc = xml.dom.minidom.parse(filename)
    top = doc.documentElement
    try:
        if top.getAttribute('uuid') == uuid:
            top.getElementsByTagName(
                "name-description")[0].firstChild.nodeValue = (config.get("update_description", uuid))
            with open(filename, "w") as fileh:
                top.writexml(fileh, addindent="  ", newl='\n')
    except ConfigParser.NoOptionError:
        return


def reduce_update_doc(in_fname, out_fname):
    """
    Remove attributes from update document to create stampfile
    """
    doc = xml.dom.minidom.parse(in_fname)
    top = doc.documentElement
    if top.hasAttribute('key'):
        top.removeAttribute('key')
    with open(out_fname, "w") as fileh:
        top.writexml(fileh, addindent="  ", newl='\n')


def edit_stampfiles(stampfiles, config, work_dir):
    """
    Replace description in rolled up update documents
    """
    for stampfile in stampfiles:
        edit_update_doc(stampfile, config, os.path.basename(stampfile))
        _, stamp_label = read_update_doc(stampfile)
        os.rename(stampfile, "%s/%s-update.xml" %
                  (os.path.dirname(stampfile), stamp_label))

    # build stampfile tarball
    stamps = os.path.join(work_dir, "stampfiles.tar")
    subprocess.check_output(['tar', '-cvf',  stamps, '.'],
                            cwd=os.path.dirname(stampfiles[0]))
    return stamps


def rebuild_update_rpm(work_dir, update_rpm, config, update_xml):
    """
    repack update RPM
    """
    stamps = None
    # extract info from update xml
    doc = xml.dom.minidom.parse(update_xml)
    top = doc.documentElement
    uuid = top.getAttribute('uuid')
    label = top.getAttribute('name-label')
    version = top.getAttribute('version')
    after_apply = top.getAttribute('after-apply-guidance')
    description = top.getElementsByTagName(
        "name-description")[0].firstChild.nodeValue
    rollsup = []
    for node in top.getElementsByTagName("rollsup"):
        rollsup.append("%s:%s" % (node.getAttribute(
            'name-label'), node.getAttribute('uuid')))
    # replace descriptions present in stampfiles
    update_rpm_dir = os.path.join(work_dir, "stamps")
    os.mkdir(update_rpm_dir)
    unpack_archive(update_rpm, update_rpm_dir)
    stampfiles = glob.glob(os.path.join(update_rpm_dir, "var/update/applied/*"))
    if stampfiles:
        stamps = edit_stampfiles(stampfiles, config, work_dir)
    local_rpms = glob.glob(os.path.join(work_dir, 'root/Packages/*.rpm'))
    pat = re.compile(r'.*/((control-)|(update-)).*')
    local_rpms = [x for x in local_rpms if not re.match(pat, x)]
    os.mkdir(os.path.join(work_dir, "SOURCES"))
    build_update_rpm(work_dir, uuid, label, version,
                     description, local_rpms, after_apply,
                     rollsup, stamps)
    new_rpm = glob.glob(os.path.join(
        work_dir, 'RPMS/noarch/update-*.rpm'))[0]
    shutil.copy2(new_rpm, os.path.join(work_dir, 'root/Packages'))


def repack_update(iso_name, args):
    """
    rebrand a single update and reassemble
    """
    work_dir = unpack_iso(iso_name)
    try:
        update_doc = os.path.join(work_dir, 'root/update.xml')
        uuid, label = read_update_doc(update_doc)
        if args.metadata:
            config = ConfigParser.ConfigParser()
            config.read(args.metadata)
            edit_update_doc(update_doc, config, uuid)
            reduce_update_doc(update_doc, os.path.join(work_dir, 'stampfile.xml'))
            rebuild_update_rpm(work_dir, glob.glob(os.path.join(
                work_dir, "root/Packages/update-%s*.rpm" % label))[0], config, update_doc)
            os.remove(os.path.join(work_dir, 'root/update.xml.asc'))
        outfile = os.path.join(args.output, os.path.basename(iso_name))
        build_iso(work_dir, outfile, uuid, label)
    finally:
        shutil.rmtree(work_dir, ignore_errors=True)


def parse_args_or_exit(argv=None):
    """
    Parse command line options
    """
    parser = argparse.ArgumentParser(description="Repack update media")
    parser.add_argument("input", metavar="ISO", nargs="+",
                        help="Input ISO file")
    parser.add_argument("-o", "--output", metavar="DIR", required=True,
                        help="Output directory")
    parser.add_argument("--metadata", metavar="FILE", required=True,
                        help="metadata file which contains keys to replace the update")

    return parser.parse_args(argv)


def main(argv):
    """
    Unpack multiple update ISOs, repack required parts then reassemble
    """
    args = parse_args_or_exit(argv)
    for iso in args.input:
        local_iso = fetch_remote_file(iso)
        repack_update(local_iso, args)
        if local_iso != iso:
            shutil.rmtree(os.path.dirname(local_iso), ignore_errors=True)


def _main():
    """
    Entry point for setuptools CLI wrapper
    """
    main(sys.argv[1:])


# Entry point when run directly
if __name__ == "__main__":
    _main()
