#!/usr/bin/env python

import argparse
import glob
import os
import os.path
import shutil
import subprocess
import sys
import tempfile
import re
import xml.dom.minidom

from update_package.common import *


def get_update_signature(work_dir):
    """
    Print pgp signature information of update ISO
    """
    # Signature status: 0 -> valid, 1 -> invalid, 2 -> nokey
    sign_status_array = ('valid', 'invalid', 'nokey')
    no_sig_string = 'Signature: None'
    doc = xml.dom.minidom.parse(os.path.join(work_dir, 'root/update.xml'))
    update_element = doc.getElementsByTagName('update')[0]
    key_file = update_element.getAttribute('key')
    if len(key_file) == 0:
        return no_sig_string

    process = subprocess.Popen(['gpg', '--verify',
                                os.path.join(work_dir, 'root/update.xml.asc')],
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
    _, gpg_err = process.communicate()

    key_id_pat = re.compile(r'.*key ID (.*)')
    key_id_match = None
    for each_line in gpg_err.split('\n'):
        key_id_match = re.match(key_id_pat, each_line)
        if key_id_match:
            break
    if key_id_match is None:
        return no_sig_string

    content = 'Signature: ' + key_id_match.group(1) + ' '
    content += key_file + \
        ' (' + sign_status_array[min(process.returncode, 2)] + ')'
    return content


def get_update_xml(work_dir):
    """
    Print XML document of update ISO
    """
    content = None
    with open(os.path.join(work_dir, 'root/update.xml')) as fileh:
        content = fileh.read()

    return 'Update.xml:\n' + content


def get_update_packages(work_dir, need_all=False):
    """
    List packages of update ISO
    """
    packages = [os.path.basename(x) for x in glob.glob(
        os.path.join(work_dir, 'root/Packages/*.rpm'))]
    if not need_all:
        pat = re.compile(r'(control-)|(update-).*')
        packages = [x for x in packages if not re.match(pat, x)]

    return 'Packages:\n' + '\n'.join(packages)


def get_rollsups(work_dir):
    """
    List rollsups update.xml of update ISO
    """
    update_rpm = glob.glob(os.path.join(
        work_dir, 'root/Packages/update-*.rpm'))
    if len(update_rpm) == 0:
        return "Rollsups: (0)"
    update_rpm_dir = os.path.join(work_dir, "stamps")
    os.mkdir(update_rpm_dir)
    unpack_archive(update_rpm[0], update_rpm_dir)
    rollsups = []
    for stampfile in glob.glob(os.path.join(update_rpm_dir, "var/update/applied/*")):
        with open(stampfile, 'r') as input_file:
            rollsups.append(input_file.read())
    return 'Rollsups: (%d)\n' % len(rollsups) + "\n".join(rollsups)


def parse_args_or_exit(argv=None):
    """
    Parse command line options
    """
    parser = argparse.ArgumentParser(description="List update info")
    parser.add_argument("update_iso", metavar="ISO",
                        help="Update ISO to check")
    parser.add_argument("-s", "--signature", action="store_true",
                        dest="print_signature", help="Print signature info")
    parser.add_argument("-x", "--xml", action="store_true",
                        dest="print_xml", help="Print XML document")
    parser.add_argument("-p", "--packages", action="store_true",
                        dest="list_packages", help="List payload packages")
    parser.add_argument("-a", "--all", action="store_true",
                        dest="all_packages", help="List all packages")
    parser.add_argument("-r", "--rollsup", action="store_true",
                        dest="list_rollsups", help="List all rollsups")
    parser.add_argument("-o", "--output", metavar="file",
                        help="Output result to file")

    args = parser.parse_args(argv)

    if args.all_packages and not args.list_packages:
        sys.exit("%s: \"--all\" only apply to \"--packages\"" % sys.argv[0])

    return args


def main(argv):
    """
    List ISO update package informations
    """
    args = parse_args_or_exit(argv)
    content = ''
    work_dir = tempfile.mkdtemp(prefix='update-info-')
    try:
        os.mkdir(os.path.join(work_dir, 'root'))
        local_update_iso = fetch_remote_file(args.update_iso, work_dir)
        unpack_archive(local_update_iso, os.path.join(work_dir, 'root'))
        if args.print_signature:
            content += get_update_signature(work_dir)
        if args.print_xml:
            content += '\n' + get_update_xml(work_dir)
        if args.list_packages:
            content += '\n' + get_update_packages(work_dir, args.all_packages)

        if args.list_rollsups:
            content += '\n' + get_rollsups(work_dir)

        content = content.lstrip('\n')
        if len(content) > 0:
            if args.output:
                with open(args.output, 'w') as fileh:
                    fileh.write(content)
            else:
                print(content)
        else:
            sys.exit("%s: Unknown error" % sys.argv[0])
    finally:
        shutil.rmtree(work_dir, ignore_errors=True)


def _main():
    """
    Entry point for setuptools CLI wrapper
    """
    main(sys.argv[1:])


# Entry point when run directly
if __name__ == "__main__":
    _main()
